/*
 * Copyright (C) 2020 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <moderncli/strings.hpp>
#include <moderncli/keyfile.hpp>
#include <moderncli/sync.hpp>

#include <unordered_set>
#include <unordered_map>
#include <atomic>

#include "driver.hpp"
#include "segments.hpp"
#include "utils.hpp"

using namespace server;
using namespace exosip;
using namespace std::literals::chrono_literals;

using map_t = std::unordered_map<int, segment_t *>;
using maplist_t = struct {
    map_t calls;
    map_t dialogs;
};

namespace {
// a lot of this is set only once at startup...
const char *AGENT_ID = SERVER_ID "/" PROJECT_VERSION;
const Session::event_t drop_event{event_type::drop};
const Session::event_t ack_event{event_type::answer};

uint16_t media_base = 0;
context_t listener(AGENT_ID);
unique_sync<maplist_t> maplist;
std::string remote_uri, route_uri, media_address;
std::thread event_thread;
std::atomic<int> active_rid = -1;
std::unordered_set<std::string> localhosts = {"localhost", "localhost.localdomain"};

[[maybe_unused]] auto is_local(osip_uri *uri) {
    if(!uri || !uri->host)
        return false;

    return std::any_of(localhosts.begin(), localhosts.end(), [uri](const std::string_view& localhost){
        return localhost == uri->host;
    });
}

auto subject(exosip::msg_t msg) {
    osip_header_t *header{nullptr};
    if(msg)
        osip_message_get_subject(msg, 0, &header);
    return (header && header->hvalue) ? std::string(header->hvalue) : "";
}

void fail_message(const event_t& event, int code) {
    const exosip::guard_t lock(event.context());
    eXosip_message_send_answer(event.context(), event.tid(), code, nullptr);
}

void fail_call(const event_t& event, int code) {
    const exosip::guard_t lock(event.context());
    eXosip_call_send_answer(event.context(), event.tid(), code, nullptr);
}

void allow_options(exosip::msg_t msg) {
    osip_message_set_header(msg, ALLOW, "OPTIONS,INVITE,BYE,CANCEL,ACK,INFO");
    osip_message_set_header(msg, ACCEPT, "application/sdp,text/plain");
    osip_message_set_header(msg, ACCEPT_ENCODING, "text/plain");
    osip_message_set_header(msg, ACCEPT_LANGUAGE, "en");    // TODO: Phrasebook
    osip_message_set_header(msg, SUPPORTED, "100rel");
}

// hidden event handlers
void do_close(const event_t& event) {
    auto cid = event.cid();
    segment_t *leg = nullptr;
    sync_ptr maps(maplist);

    try {
        leg = maps->calls.at(cid);
    }
    catch(...) {
        debug(5, "no call leg for cid={}", cid);
        return;
    }

    // Ignore if no active call
    if(leg->cid < 1)
        return;

    debug(4, "closing call for cid={}", cid);
    auto sid = leg->sid();
    maps->calls.erase(leg->cid);
    if(leg->did > 0)
        maps->dialogs.erase(leg->did);
    leg->cid = leg->did = leg->tid = 0;
    maps.unlock();

    try {
        auto session = Driver::get(sid);
        Session::send(session, drop_event);
    }
    catch(...) {}
}

auto do_ack(const event_t& event) {
    auto request = event.request();
    auto cid = event.cid();
    auto did = event.did();
    segment_t *leg = nullptr;
    sync_ptr maps(maplist);

    try {
        leg = maps->calls.at(cid);
    }
    catch(...) {
        debug(5, "no call leg for cid={}", cid);
        return;
    }

    // if no rtp or call, we can ignore
    if(leg->cid < 1)
        return;

    debug(4, "ack call for cid={}", cid);
    auto sid = leg->sid();
    if(did > 0) {
        if(did != leg->did) {
            if(leg->did > 0)
                maps->dialogs.erase(leg->did);
            maps->dialogs[did] = leg;
        }
        leg->did = did;
        leg->tid = -1;
    }

    std::string_view sdp;
    osip_body_t *body{nullptr};
    osip_message_get_body(request, 0, &body);
    if(body && body->body && body->length)
        sdp = std::string_view(body->body, body->length);

    maps.unlock();
    leg->apply_sdp(sdp);
    // TODO: start rtp session...

    try {
        auto session = Driver::get(sid);
        Session::send(session, ack_event);
    }
    catch(...) {}
}

auto do_invite(const event_t& event) {
    auto request = event.request();
    auto from = request->from;
    auto to = request->to;
    auto caller_from = std::string(from->url->username ? from->url->username : "unknown");
    auto caller_id = unquote<std::string>(from->displayname ? from->displayname : "unknown");
    auto dialed_to = std::string(to->url->username ? to->url->username : "unknown");
    auto start = "main";
    script scr{};   // NOLINT

    try {
        scr = Driver::get_script("default.scr");
    }
    catch(...) {}

    try {
        if(!scr)
            scr = Driver::get_script("generic.scr");
    }
    catch(...) {}

    try {
        if(dialed_to != net::uri_userid(remote_uri))
            scr = Driver::get_script(dialed_to + ".scr");
        scr = Driver::get_script(caller_from + ".scr");
    }
    catch(...) {}

    if(scr.empty()) {
        log_error("cannot find script for call from {} to {}", caller_from, dialed_to);
        return SIP_NOT_FOUND;
    }

    log_notice("call from \"{}\" at {} to {}\n", caller_id, caller_from, dialed_to);

    // ignite new call leg
    log_notice("call from \"{}\" at {} to {}\n", caller_id, caller_from, dialed_to);
    const session_t session(std::make_shared<segment_t>(scr, event));
    auto leg = dynamic_cast<segment_t*>(session.get());
    leg->make_const({
        {"SUBJECT", subject(request)},
        {"CALLER", caller_from},
        {"DIALED", dialed_to},
        {"FROM", caller_from},
        {"TO", dialed_to},
        {"CALLER_ID", caller_id},
    });

    if(!Driver::attach(session))
        return SIP_TEMPORARILY_UNAVAILABLE;

    auto port = media_base + (2 * leg->slot());
    auto ipc = leg->ipc();
    std::string_view sdp{}; // NOLINT
    osip_body_t *body{nullptr};
    osip_message_get_body(request, 0, &body);
    if(body && body->body && body->length)
        sdp = std::string_view(body->body, body->length);

    str_copy(ipc->caller, sizeof(ipc->caller), caller_from);
    str_copy(ipc->dialed, sizeof(ipc->dialed), dialed_to);
    leg->ipc()->state = ipc_session::ISC_INVITING;
    leg->ipc()->type = ipc_session::ISC_INBOUND;
    leg->connect(media_address, sdp, port);
    if(leg->is_label("ring")) {
        start = "ring";
        if(!leg->answer(SIP_RINGING))
            return SIP_TEMPORARILY_UNAVAILABLE;
        leg->ipc()->state = ipc_session::ISC_RINGING;
    }
    else if(leg->is_label("main")) {
        if(!leg->answer(SIP_OK, [leg](auto msg) {
            auto sdp = leg->sdp();
            if(!sdp.empty()) {
                osip_message_set_body(msg, sdp.data(), sdp.size());
                osip_message_set_content_type(msg, "application/sdp");
            }
        }))
            return SIP_TEMPORARILY_UNAVAILABLE;   // TODO: ANSWER, so we need sdp from rtp...
        leg->stream();
    }
    else
        return SIP_TEMPORARILY_UNAVAILABLE;

    sync_ptr maps(maplist); // NOLINT
    maps->calls[leg->cid] = leg;
    if(leg->did > 0)
        maps->dialogs[leg->did] = leg;

    Session::begin(session, start);
    return SIP_OK;
}

void do_options(const event_t& event) {
    auto context = event.context();
    auto user = event.request()->to->url->username; // to->url proven earlier
    auto tid = event.tid();

    if(!user || eq(user, "system")) {   // Ping!
        auto msg = exosip::msg_t(nullptr);
        const exosip::guard_t lock(context);
        eXosip_options_build_answer(context, tid, SIP_OK, &msg);
        allow_options(msg);
        eXosip_options_send_answer(context, tid, SIP_OK, msg);
    }
    else {
        const exosip::guard_t lock(context);
        eXosip_options_send_answer(context, tid, SIP_FORBIDDEN, nullptr);
    }
}
} // end namespace

driver driver::instance;

driver::driver() : Driver("generic") {}

void driver::configure(keyfile& config) {
    const auto& server = config["server"];
    const auto& hostname = server.at("hostname");

    auto& keys = config["generic"];
    auto& identity = keys["identity"];
    auto& address = keys["address"];
    auto& port_id = keys["port"];   // cppcheck-suppress variableScope
    auto& authorize = keys["authorize"];
    auto& password = keys["password"];
    auto& userid = keys["userid"];
    auto& route = keys["route"];
    auto& host = keys["host"];
    auto& type = keys["type"];

    auto env = util::get_env("BORDEAUX_IDENTITY", 255);
    if(env && !strncmp(env, "sip:", 4) && strchr(env, '@'))
        identity = env + 4;
    else if(env)
        identity = env;

    if(!identity.empty()) {
        if(identity.substr(0, 4) == "sip:")
            identity = identity.substr(4);
        auto pos = identity.find_last_of('@');
        if(pos != std::string::npos) {
            authorize = identity.substr(0, pos);
            host = identity.substr(pos + 1);
            pos = authorize.find_first_of(':');
            identity = "sip:" + authorize.substr(0, pos) + "@" + host;
        }
        else {
            authorize = identity;
            identity.clear();
            pos = authorize.find_first_of(':');
        }

        if(pos != std::string::npos) {
            password = authorize.substr(pos + 1);
            authorize = authorize.substr(0, pos);
        }

        if(userid.empty())
            userid = authorize;
    }

    std::string uri;
    env = util::get_env("BORDEAUX_DRIVER", 256);
    if(env)
        uri = env;

    if(!uri.empty() && uri.substr(0, 4) == "sip:") {
        uri = uri.substr(4);
        auto pos = uri.find_last_of(':');
        if(pos != std::string::npos) {
            port_id = uri.substr(pos + 1);
            address = uri.substr(0, pos);
        }
        else {
            port_id = "5060";
            address = uri;
        }
    }

    if(host.empty())
        host = hostname;

    // authorization id
    if(authorize.empty())
        userid = server.at("username");

    // if no separate authorize maybe user...
    if(authorize.empty())
        authorize = userid;

    auto route_port = net::uri_port(host);

    if(route.empty())
        route = net::make_uri("sip:", net::uri_host(host), route_port);
    else if(route == "static")
        route = net::make_uri("sip:", net::uri_address(host), route_port);
    else {
        auto route_host = net::uri_host(route);
        route_port = net::uri_port(route);
        route = net::make_uri("sip:", route_host, route_port);
    }

    if(identity.empty())
        identity = "sip:" + authorize + "@" + net::uri_host(host);

    // none means no registration
    if(userid.empty())
        userid = "none";

    // cannot authorize if no credentials
    if(password.empty() || authorize.empty())
        authorize = "none";

    if(!identity.empty()) {
        auto pos = identity.find_first_of('@');
        if(pos == std::string::npos)
            identity = userid + "@" + identity;
    }

    if(!begins_with(identity, "sip:"))
        identity = "sip:" + identity;

    if(address.empty())
        address = "*";

    if(type.empty())
        type = "udp";

    Driver::configure(config);
}

void driver::restart(const keyfile& config) {
    static auto prior_registration = "none"s;

    const auto& keys = config["generic"];
    auto identity = keys.at("identity").c_str();
    auto route = keys.at("route").c_str();
    auto username = keys.at("userid").c_str();
    auto authorize = keys.at("authorize").c_str();
    auto password = keys.at("password").c_str();
    auto expires = 300;     // default 5 minutes

    try {
        expires = stoi(keys.at("expires"));
    }
    catch(...) {
    }

    log_info("remote identity {}", identity);
    log_info("remote routing {}", route);
    log_info("registers {}, authorizes {}", username, authorize);

    const exosip::guard_t lock(listener); // lock events while updating
    route_uri = route;
    remote_uri = identity;
    Driver::restart(config);        // kick off scripts
    eXosip_clear_authentication_info(*listener);
    if(password && *password && authorize != "none"s)
        eXosip_add_authentication_info(*listener, username, authorize, password, nullptr, nullptr);

    if(username != prior_registration) {
        eXosip_unregister();
        auto msg = exosip::msg_t(nullptr);
        active_rid = eXosip_register_build_initial_register(*listener, identity, route, nullptr, expires, &msg);
        if(active_rid < 0) {
            log_error("failed to register {}", remote_uri);
            return;
        }

        log_info("starting registration {}", remote_uri);
        allow_options(msg);
        eXosip_register_send_register(*listener, active_rid, msg);
        prior_registration = username;
    }
}

void driver::startup(const keyfile& config) {
    const auto& server = config["server"];
    const auto& keys = config["generic"];
    const auto& hostname = server.at("hostname");
    const auto& port_id = keys.at("port");

    auto address = keys.at("address");
    auto type = listen_type::udp;
    auto port = 0;

    if(keys.at("type") == "tcp")
        type = listen_type::tcp;
    else if(keys.at("type") == "tls")
        type = listen_type::tls;

    try {
        address = net::if_address(address);
    }
    catch(const std::exception&) {
        log_fail(98, "{}: driver address invalid", address);
    }

    try {
        port = stoi(port_id);
    }
    catch(...) {
    }

    if(!port) {
        port = listener.find_port(5062,
            type == listen_type::udp ? IPPROTO_UDP : IPPROTO_TCP);
    }

    listener.set_family(ipv6 ? family_type::ipv6 : family_type::ipv4);
    listener.srv_lookup(false);

    if(!listener.listen(address, type, port))
        log_fail(98, "cannot bind to {}:{}", address, port);

    log_info("starting {}", net::to_string(address, static_cast<unsigned short>(port)));

    media_address = server.at("media");
    media_base = port + 2;
    localhosts.emplace(ipv6 ? "[::1]" : "127.0.0.1");
    localhosts.emplace(media_address);
    localhosts.emplace(hostname);
    localhosts.emplace(net::uri_address(hostname));
    localhosts.emplace(net::hostname());
    localhosts.emplace(net::uri_address(net::hostname()));
    log_info("media address {}", media_address);

    restart(config);    // kick off scripts, etc
    event_thread = std::thread(&driver::dispatch, this, 1);
}

void driver::disconnect() {
    if(listener.active() && active_rid != -1) {
        const exosip::guard_t lock(*listener);
        eXosip_unregister();
    }
    if(listener.active()) {
        Driver::disconnect();
        std::this_thread::sleep_for(std::chrono::milliseconds(960));
    }
    else
        Driver::disconnect();
}

void driver::shutdown() {
    sync_ptr maps(maplist);
    for(auto [cid, leg] : maps->calls) {
        leg->terminate();
    }

    Driver::shutdown();

    if(listener.active()) {
        listener.close();
        event_thread.join();
    }
}

void driver::answer(Session *session) {
    auto leg = dynamic_cast<segment_t*>(session);
    leg->answer(SIP_OK, [leg](auto msg) {
        auto sdp = leg->sdp();
        if(!sdp.empty()) {
            osip_message_set_body(msg, sdp.data(), sdp.size());
                osip_message_set_content_type(msg, "application/sdp");
        }
    });
}

void driver::drop(Session *session) {
    auto leg = dynamic_cast<segment_t*>(session);
    leg->disconnect();

    sync_ptr maps(maplist);
    if(leg->did > 0)
        maps->dialogs.erase(leg->did);
    if(leg->cid > 0)
        maps->calls.erase(leg->cid);
    leg->terminate();
    debug(4, "map counts {}, {}\n", maps->dialogs.size(), maps->calls.size());
}

void driver::eXosip_unregister() {
    if(active_rid == -1)
        return;

    debug(3, "releasing registration");

    // Locked before we get here...
    auto msg = exosip::msg_t(nullptr);
    auto res = eXosip_register_build_register(*listener, active_rid, 0, &msg);
    if(res > -1)
        eXosip_register_send_register(*listener, active_rid, msg);
    else
        log_error("unregister failed");
    active_rid = -1;
    Driver::notify(false);
    sys_mapped->active = false;
}

// event dispatch listener thread
void driver::dispatch(int priority) {
    if(!util::set_priority(priority))
        priority = 0;

    log_info("starting listener thread; pri={}", priority);

    auto prior = sync_clock();
    for(;;) {
        const exosip::event_t event(listener, 500);
        if(!running)
            break;

        // breakdown events to dispatch handlers with minimal stack frames
        if(is(event)) {
            debug(5, "event: {}", event.text());

            auto request = event.request();
            auto response = event.response();
            auto status_code = response ? response->status_code : 409;

            switch(event.type()) {
            case EXOSIP_MESSAGE_NEW:
                if(!request)
                    fail_message(event, SIP_BAD_REQUEST);
                else if(MSG_IS_OPTIONS(request))
                    do_options(event);
                else
                    fail_message(event, SIP_METHOD_NOT_ALLOWED);
                break;
            case EXOSIP_CALL_INVITE:
                if(!request || event.cid() < 1) {
                    debug(2, "bad invite request");
                    fail_call(event, SIP_BAD_REQUEST);
                }
                else if(sys_mapped->used >= limit_) {
                    log_warn("declined call, all sessions busy");
                    fail_call(event, SIP_BUSY_HERE);
                }
                else {
                    status_code = do_invite(event);
                    if(status_code != SIP_OK) {
                        log_error("incoming call failed, code={}", status_code);
                        fail_call(event, status_code);
                    }
                }
                break;
            case EXOSIP_CALL_MESSAGE_NEW:
                if(!request || event.cid() < 1) {
                    debug(2, "bad invite request");
                    fail_call(event, SIP_BAD_REQUEST);
                }
                else if(MSG_IS_BYE(request)) {
                    do_close(event);
                }
                else
                    fail_call(event, SIP_METHOD_NOT_ALLOWED);
                break;
            case EXOSIP_REGISTRATION_FAILURE:
                if(active_rid != -1 && (status_code == SIP_UNAUTHORIZED || status_code == SIP_PROXY_AUTHENTICATION_REQUIRED)) {
                    debug(1, "send credentials");
                    event.automatic();
                }
                else {
                    Driver::notify(false);
                    sys_mapped->active = false;
                    log_warn("registration failed; code={}", status_code);
                }
                break;
            case EXOSIP_REGISTRATION_SUCCESS:
                if(active_rid == -1)
                    log_info("released registration");
                else if(!online) {
                    Driver::notify(true);
                    sys_mapped->active = true;
                    log_info("registered with server");
                }
                else
                    debug(1, "registration confirmed");
                break;
            case EXOSIP_CALL_RELEASED:
            case EXOSIP_CALL_CLOSED:
                do_close(event);
                break;
            case EXOSIP_CALL_ACK:
                do_ack(event);
                break;
            default:
                event.automatic();
            }
        }

        auto current = sync_clock();
        if(!event || current > prior + 500ms) {	        // timeout
            debug(6, "listener automatic; online={}", static_cast<int>(online));
            prior = current;
            listener.automatic();
            continue;
        }
    }
    log_info("stopping listener thread");
}

driver& server::sip = driver::instance;
